@@ -19,11 +19,12 @@
     Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
 */
 #include <errno.h>
+#include <poll.h>
 #include "sd_bus_internals.h"
 
 static void SdBus_dealloc(SdBusObject* self) {
         sd_bus_unref(self->sd_bus_ref);
-        Py_XDECREF(self->reader_fd);
+        Py_XDECREF(self->bus_fd);
 
         SD_BUS_DEALLOC_TAIL;
 }
@@ -229,46 +230,24 @@ int future_set_exception_from_message(PyObject* future, sd_bus_message* message)
         return 0;
 }
 
-static PyObject* SdBus_drive(SdBusObject* self, PyObject* Py_UNUSED(args));
-
 static PyObject* SdBus_get_fd(SdBusObject* self, PyObject* Py_UNUSED(args)) {
         int file_descriptor = CALL_SD_BUS_AND_CHECK(sd_bus_get_fd(self->sd_bus_ref));
 
         return PyLong_FromLong((long)file_descriptor);
 }
 
-#define CHECK_SD_BUS_READER                                             \
-        ({                                                              \
-                if (self->reader_fd == NULL) {                          \
-                        CALL_PYTHON_EXPECT_NONE(register_reader(self)); \
-                }                                                       \
-        })
+static PyObject* SdBus_asyncio_update_fd_watchers(SdBusObject* self);
 
-PyObject* register_reader(SdBusObject* self) {
-        PyObject* running_loop CLEANUP_PY_OBJECT = CALL_PYTHON_AND_CHECK(PyObject_CallFunctionObjArgs(asyncio_get_running_loop, NULL));
-        PyObject* new_reader_fd CLEANUP_PY_OBJECT = CALL_PYTHON_AND_CHECK(SdBus_get_fd(self, NULL));
-        PyObject* drive_method CLEANUP_PY_OBJECT = CALL_PYTHON_AND_CHECK(PyObject_GetAttrString((PyObject*)self, "drive"));
-        Py_XDECREF(CALL_PYTHON_AND_CHECK(PyObject_CallMethodObjArgs(running_loop, add_reader_str, new_reader_fd, drive_method, NULL)));
-        Py_INCREF(new_reader_fd);
-        self->reader_fd = new_reader_fd;
-        Py_RETURN_NONE;
-}
+#define CHECK_ASYNCIO_WATCHERS ({ CALL_PYTHON_EXPECT_NONE(SdBus_asyncio_update_fd_watchers(self)); })
 
-PyObject* unregister_reader(SdBusObject* self) {
-        PyObject* running_loop CLEANUP_PY_OBJECT = CALL_PYTHON_AND_CHECK(PyObject_CallFunctionObjArgs(asyncio_get_running_loop, NULL));
-        Py_XDECREF(CALL_PYTHON_AND_CHECK(PyObject_CallMethodObjArgs(running_loop, remove_reader_str, self->reader_fd, NULL)));
-        Py_RETURN_NONE;
-}
-
-static PyObject* SdBus_drive(SdBusObject* self, PyObject* Py_UNUSED(args)) {
+static PyObject* SdBus_process(SdBusObject* self, PyObject* Py_UNUSED(args)) {
         int return_value = 1;
         while (return_value > 0) {
                 return_value = sd_bus_process(self->sd_bus_ref, NULL);
                 if (return_value < 0) {
-                        CALL_PYTHON_AND_CHECK(unregister_reader(self));
                         if (-ECONNRESET == return_value) {
                                 // Connection gracefully terminated
-                                Py_RETURN_NONE;
+                                break;
                         } else {
                                 // Error occurred processing sdbus
                                 CALL_SD_BUS_AND_CHECK(return_value);
@@ -280,6 +259,7 @@ static PyObject* SdBus_drive(SdBusObject* self, PyObject* Py_UNUSED(args)) {
                         return NULL;
                 }
         }
+        CHECK_ASYNCIO_WATCHERS;
 
         Py_RETURN_NONE;
 }
@@ -291,7 +271,7 @@ int SdBus_async_callback(sd_bus_message* m,
         PyObject* py_future = userdata;
         PyObject* is_cancelled CLEANUP_PY_OBJECT = PyObject_CallMethod(py_future, "cancelled", "");
         if (Py_True == is_cancelled) {
-                // A bit unpythonic but SdBus_drive does not error out
+                // A bit unpythonic but SdBus_process does not error out
                 return 0;
         }
 
@@ -340,7 +320,7 @@ static PyObject* SdBus_call_async(SdBusObject* self, PyObject* args) {
         if (PyObject_SetAttrString(new_future, "_sd_bus_py_slot", (PyObject*)new_slot_object) < 0) {
                 return NULL;
         }
-        CHECK_SD_BUS_READER;
+        CHECK_ASYNCIO_WATCHERS;
         return new_future;
 }
 
@@ -376,17 +356,16 @@ static PyObject* SdBus_add_interface(SdBusObject* self, PyObject* args) {
 }
 
 int _SdBus_signal_callback(sd_bus_message* m, void* userdata, sd_bus_error* Py_UNUSED(ret_error)) {
-        PyObject* async_queue = userdata;
+        PyObject* signal_callback = userdata;
 
-        SdBusMessageObject* new_message_object CLEANUP_SD_BUS_MESSAGE = (SdBusMessageObject*)SD_BUS_PY_CLASS_DUNDER_NEW(SdBusMessage_class);
-        if (new_message_object == NULL) {
-                return -1;
-        }
+        PyObject* running_loop CLEANUP_PY_OBJECT = CALL_PYTHON_CHECK_RETURN_NEG1(PyObject_CallFunctionObjArgs(asyncio_get_running_loop, NULL));
+
+        SdBusMessageObject* new_message_object CLEANUP_SD_BUS_MESSAGE =
+            (SdBusMessageObject*)CALL_PYTHON_CHECK_RETURN_NEG1(SD_BUS_PY_CLASS_DUNDER_NEW(SdBusMessage_class));
         _SdBusMessage_set_messsage(new_message_object, m);
-        PyObject* should_be_none CLEANUP_PY_OBJECT = PyObject_CallMethodObjArgs(async_queue, put_no_wait_str, new_message_object, NULL);
-        if (should_be_none == NULL) {
-                return -1;
-        }
+
+        Py_XDECREF(CALL_PYTHON_CHECK_RETURN_NEG1(PyObject_CallMethodObjArgs(running_loop, call_soon_str, signal_callback, new_message_object, NULL)));
+
         return 0;
 }
 
@@ -394,21 +373,15 @@ int _SdBus_match_signal_instant_callback(sd_bus_message* m, void* userdata, sd_b
         PyObject* new_future = userdata;
 
         if (!sd_bus_message_is_method_error(m, NULL)) {
-                PyObject* new_queue CLEANUP_PY_OBJECT = PyObject_GetAttrString(new_future, "_sd_bus_queue");
-                if (new_queue == NULL) {
-                        return -1;
-                }
+                SdBusSlotObject* slot_object CLEANUP_SD_BUS_SLOT =
+                    (SdBusSlotObject*)CALL_PYTHON_CHECK_RETURN_NEG1(PyObject_GetAttrString(new_future, "_sd_bus_slot"));
 
-                PyObject* should_be_none CLEANUP_PY_OBJECT = PyObject_CallMethodObjArgs(new_future, set_result_str, new_queue, NULL);
-                if (should_be_none == NULL) {
-                        return -1;
-                }
+                Py_XDECREF(CALL_PYTHON_CHECK_RETURN_NEG1(PyObject_CallMethodObjArgs(new_future, set_result_str, slot_object, NULL)));
 
-                SdBusSlotObject* slot_object CLEANUP_SD_BUS_SLOT = (SdBusSlotObject*)PyObject_GetAttrString(new_queue, "_sd_bus_slot");
-                if (slot_object == NULL) {
-                        return -1;
-                }
-                sd_bus_slot_set_userdata(slot_object->slot_ref, new_queue);
+                PyObject* signal_callback = CALL_PYTHON_CHECK_RETURN_NEG1(PyObject_GetAttrString(new_future, "_sd_bus_signal_callback"));
+
+                sd_bus_slot_set_userdata(slot_object->slot_ref, signal_callback);
+                sd_bus_slot_set_destroy_callback(slot_object->slot_ref, (sd_bus_destroy_t)Py_DecRef);
         } else {
                 if (future_set_exception_from_message(new_future, m) < 0) {
                         return -1;
@@ -424,46 +397,44 @@ static int _unicode_or_none(PyObject* some_object) {
         return (PyUnicode_Check(some_object) || (Py_None == some_object));
 }
 
-static PyObject* SdBus_get_signal_queue(SdBusObject* self, PyObject* const* args, Py_ssize_t nargs) {
-        SD_BUS_PY_CHECK_ARGS_NUMBER(4);
+static PyObject* SdBus_match_signal_async(SdBusObject* self, PyObject* const* args, Py_ssize_t nargs) {
+        SD_BUS_PY_CHECK_ARGS_NUMBER(5);
 
         SD_BUS_PY_CHECK_ARG_CHECK_FUNC(0, _unicode_or_none);
         SD_BUS_PY_CHECK_ARG_CHECK_FUNC(1, _unicode_or_none);
         SD_BUS_PY_CHECK_ARG_CHECK_FUNC(2, _unicode_or_none);
         SD_BUS_PY_CHECK_ARG_CHECK_FUNC(3, _unicode_or_none);
+        SD_BUS_PY_CHECK_ARG_CHECK_FUNC(4, PyCallable_Check);
 
         const char* sender_service_char_ptr = SD_BUS_PY_UNICODE_AS_CHAR_PTR_OPTIONAL(args[0]);
         const char* path_name_char_ptr = SD_BUS_PY_UNICODE_AS_CHAR_PTR_OPTIONAL(args[1]);
         const char* interface_name_char_ptr = SD_BUS_PY_UNICODE_AS_CHAR_PTR_OPTIONAL(args[2]);
         const char* member_name_char_ptr = SD_BUS_PY_UNICODE_AS_CHAR_PTR_OPTIONAL(args[3]);
+        PyObject* signal_callback = args[4];
 #else
-static PyObject* SdBus_get_signal_queue(SdBusObject* self, PyObject* args) {
+static PyObject* SdBus_match_signal_async(SdBusObject* self, PyObject* args) {
         const char* sender_service_char_ptr = NULL;
         const char* path_name_char_ptr = NULL;
         const char* interface_name_char_ptr = NULL;
         const char* member_name_char_ptr = NULL;
-        CALL_PYTHON_BOOL_CHECK(
-            PyArg_ParseTuple(args, "zzzz", &sender_service_char_ptr, &path_name_char_ptr, &interface_name_char_ptr, &member_name_char_ptr, NULL));
+        PyObject* signal_callback = NULL;
+        CALL_PYTHON_BOOL_CHECK(PyArg_ParseTuple(args, "zzzzO", &sender_service_char_ptr, &path_name_char_ptr, &interface_name_char_ptr, &member_name_char_ptr,
+                                                &signal_callback, NULL));
 #endif
-        SdBusSlotObject* new_slot CLEANUP_SD_BUS_SLOT = (SdBusSlotObject*)CALL_PYTHON_AND_CHECK(SD_BUS_PY_CLASS_DUNDER_NEW(SdBusSlot_class));
-
-        PyObject* new_queue CLEANUP_PY_OBJECT = CALL_PYTHON_AND_CHECK(PyObject_CallFunctionObjArgs(asyncio_queue_class, NULL));
-
-        // Bind lifetime of the slot to the queue
-        CALL_PYTHON_INT_CHECK(PyObject_SetAttrString(new_queue, "_sd_bus_slot", (PyObject*)new_slot));
-
         PyObject* running_loop CLEANUP_PY_OBJECT = CALL_PYTHON_AND_CHECK(PyObject_CallFunctionObjArgs(asyncio_get_running_loop, NULL));
-
         PyObject* new_future CLEANUP_PY_OBJECT = CALL_PYTHON_AND_CHECK(PyObject_CallMethod(running_loop, "create_future", ""));
 
-        // Bind lifetime of the queue to future
-        CALL_PYTHON_INT_CHECK(PyObject_SetAttrString(new_future, "_sd_bus_queue", new_queue));
+        SdBusSlotObject* new_slot CLEANUP_SD_BUS_SLOT = (SdBusSlotObject*)CALL_PYTHON_AND_CHECK(SD_BUS_PY_CLASS_DUNDER_NEW(SdBusSlot_class));
+
+        // Bind lifetime of the slot to the Future
+        CALL_PYTHON_INT_CHECK(PyObject_SetAttrString(new_future, "_sd_bus_slot", (PyObject*)new_slot));
+        CALL_PYTHON_INT_CHECK(PyObject_SetAttrString(new_future, "_sd_bus_signal_callback", signal_callback));
 
         CALL_SD_BUS_AND_CHECK(sd_bus_match_signal_async(self->sd_bus_ref, &new_slot->slot_ref, sender_service_char_ptr, path_name_char_ptr,
                                                         interface_name_char_ptr, member_name_char_ptr, _SdBus_signal_callback,
                                                         _SdBus_match_signal_instant_callback, new_future));
 
-        CHECK_SD_BUS_READER;
+        CHECK_ASYNCIO_WATCHERS;
         Py_INCREF(new_future);
         return new_future;
 }
@@ -474,7 +445,7 @@ int SdBus_request_name_callback(sd_bus_message* m,
         PyObject* py_future = userdata;
         PyObject* is_cancelled CLEANUP_PY_OBJECT = PyObject_CallMethod(py_future, "cancelled", "");
         if (Py_True == is_cancelled) {
-                // A bit unpythonic but SdBus_drive does not error out
+                // A bit unpythonic but SdBus_process does not error out
                 return 0;
         }
 
@@ -540,7 +511,7 @@ static PyObject* SdBus_request_name_async(SdBusObject* self, PyObject* args) {
             sd_bus_request_name_async(self->sd_bus_ref, &new_slot_object->slot_ref, service_name_char_ptr, flags, SdBus_request_name_callback, new_future));
 
         CALL_PYTHON_INT_CHECK(PyObject_SetAttrString(new_future, "_sd_bus_py_slot", (PyObject*)new_slot_object));
-        CHECK_SD_BUS_READER;
+        CHECK_ASYNCIO_WATCHERS;
         return new_future;
 }
 
@@ -644,18 +615,57 @@ static PyObject* SdBus_start(SdBusObject* self, PyObject* Py_UNUSED(args)) {
         Py_RETURN_NONE;
 }
 
+static inline int sd_bus_get_events_zero_on_closed(SdBusObject* self) {
+        int events = sd_bus_get_events(self->sd_bus_ref);
+        if (-ENOTCONN == events) {
+                return 0;
+        }
+        return events;
+};
+
+static PyObject* SdBus_asyncio_update_fd_watchers(SdBusObject* self) {
+        int events_to_watch = CALL_SD_BUS_AND_CHECK(sd_bus_get_events_zero_on_closed(self));
+        if (events_to_watch == self->asyncio_watchers_last_state) {
+                // Do not update the watchers because state is the same
+                Py_RETURN_NONE;
+        } else {
+                self->asyncio_watchers_last_state = events_to_watch;
+        }
+
+        PyObject* running_loop CLEANUP_PY_OBJECT = CALL_PYTHON_AND_CHECK(PyObject_CallFunctionObjArgs(asyncio_get_running_loop, NULL));
+        PyObject* drive_method CLEANUP_PY_OBJECT = CALL_PYTHON_AND_CHECK(PyObject_GetAttrString((PyObject*)self, "process"));
+
+        if (NULL == self->bus_fd) {
+                self->bus_fd = CALL_PYTHON_AND_CHECK(SdBus_get_fd(self, NULL));
+        }
+
+        if (events_to_watch & POLLIN) {
+                Py_XDECREF(CALL_PYTHON_AND_CHECK(PyObject_CallMethodObjArgs(running_loop, add_reader_str, self->bus_fd, drive_method, NULL)));
+        } else {
+                Py_XDECREF(CALL_PYTHON_AND_CHECK(PyObject_CallMethodObjArgs(running_loop, remove_reader_str, self->bus_fd, NULL)));
+        }
+
+        if (events_to_watch & POLLOUT) {
+                Py_XDECREF(CALL_PYTHON_AND_CHECK(PyObject_CallMethodObjArgs(running_loop, add_writer_str, self->bus_fd, drive_method, NULL)));
+        } else {
+                Py_XDECREF(CALL_PYTHON_AND_CHECK(PyObject_CallMethodObjArgs(running_loop, remove_writer_str, self->bus_fd, NULL)));
+        }
+
+        Py_RETURN_NONE;
+}
+
 static PyMethodDef SdBus_methods[] = {
     {"call", (SD_BUS_PY_FUNC_TYPE)SdBus_call, SD_BUS_PY_METH, PyDoc_STR("Send message and block until the reply.")},
     {"call_async", (SD_BUS_PY_FUNC_TYPE)SdBus_call_async, SD_BUS_PY_METH, PyDoc_STR("Async send message, returns awaitable future.")},
-    {"drive", (PyCFunction)SdBus_drive, METH_NOARGS, PyDoc_STR("Drive connection.")},
+    {"process", (PyCFunction)SdBus_process, METH_NOARGS, PyDoc_STR("Process pending IO work.")},
     {"get_fd", (SD_BUS_PY_FUNC_TYPE)SdBus_get_fd, SD_BUS_PY_METH, PyDoc_STR("Get file descriptor to poll on.")},
     {"new_method_call_message", (SD_BUS_PY_FUNC_TYPE)SdBus_new_method_call_message, SD_BUS_PY_METH, PyDoc_STR("Create new empty method call message.")},
     {"new_property_get_message", (SD_BUS_PY_FUNC_TYPE)SdBus_new_property_get_message, SD_BUS_PY_METH, PyDoc_STR("Create new empty property get message.")},
     {"new_property_set_message", (SD_BUS_PY_FUNC_TYPE)SdBus_new_property_set_message, SD_BUS_PY_METH, PyDoc_STR("Create new empty property set message.")},
     {"new_signal_message", (SD_BUS_PY_FUNC_TYPE)SdBus_new_signal_message, SD_BUS_PY_METH, PyDoc_STR("Create new empty signal message.")},
     {"add_interface", (SD_BUS_PY_FUNC_TYPE)SdBus_add_interface, SD_BUS_PY_METH, PyDoc_STR("Add interface to the bus.")},
-    {"get_signal_queue_async", (SD_BUS_PY_FUNC_TYPE)SdBus_get_signal_queue, SD_BUS_PY_METH,
-     PyDoc_STR("Returns a future that returns a queue that queues signal messages.")},
+    {"match_signal_async", (SD_BUS_PY_FUNC_TYPE)SdBus_match_signal_async, SD_BUS_PY_METH,
+     PyDoc_STR("Register signal callback asynchronously. Returns a Future that returns a SdBusSlot.")},
     {"request_name_async", (SD_BUS_PY_FUNC_TYPE)SdBus_request_name_async, SD_BUS_PY_METH, PyDoc_STR("Request D-Bus name async.")},
     {"request_name", (SD_BUS_PY_FUNC_TYPE)SdBus_request_name, SD_BUS_PY_METH, PyDoc_STR("Request D-Bus name blocking.")},
     {"add_object_manager", (SD_BUS_PY_FUNC_TYPE)SdBus_add_object_manager, SD_BUS_PY_METH, PyDoc_STR("Add object manager at the path.")},
@@ -678,8 +688,31 @@ static PyObject* SdBus_address_getter(SdBusObject* self, void* Py_UNUSED(closure
         return PyUnicode_FromString(bus_address);
 }
 
+static PyObject* SdBus_method_call_timeout_usec_getter(SdBusObject* self, void* Py_UNUSED(closure)) {
+        uint64_t timeout_usec = 0;
+        CALL_SD_BUS_AND_CHECK(sd_bus_get_method_call_timeout(self->sd_bus_ref, &timeout_usec));
+
+        return PyLong_FromUnsignedLongLong((unsigned long long)timeout_usec);
+}
+
+static int SdBus_method_call_timeout_usec_setter(SdBusObject* self, PyObject* new_value, void* Py_UNUSED(closure)) {
+        if (NULL == new_value) {
+                PyErr_SetString(PyExc_ValueError, "Cannot delete method call timeout value");
+                return -1;
+        }
+
+        unsigned long long new_timeout_usec = PyLong_AsUnsignedLongLong(new_value);
+        if ((((unsigned long long)-1) == new_timeout_usec) && (PyErr_Occurred() != NULL)) {
+                return -1;
+        }
+        CALL_SD_BUS_CHECK_RETURN_NEG1(sd_bus_set_method_call_timeout(self->sd_bus_ref, (uint64_t)new_timeout_usec));
+        return 0;
+}
+
 static PyGetSetDef SdBus_properies[] = {
     {"address", (getter)SdBus_address_getter, NULL, PyDoc_STR("Bus address."), NULL},
+    {"method_call_timeout_usec", (getter)SdBus_method_call_timeout_usec_getter, (setter)SdBus_method_call_timeout_usec_setter,
+     PyDoc_STR("D-Bus call timeout in microseconds."), NULL},
     {0},
 };
 
